import datetime
import json
import os
import subprocess
import time

from .parser import Parser


class Server:
    def __init__(self, jar_path: str, min_RAM: str = '1G', max_RAM: str = '3G', parser=None, update_wait: int = 3):
        if parser is None:
            parser = Parser()

        self.update_wait = update_wait
        self.parser = parser
        self.max_RAM = max_RAM
        self.min_RAM = min_RAM
        self.jar = jar_path

        self._log_file_name = 'temp-log.txt'

        self._server = None
        self._old_events = 0

        self.online = False

        self.abs_cwd, self.jar = os.path.split(self.jar)
        if not os.path.isabs(self.abs_cwd):
            self.abs_cwd = os.path.join(os.getcwd(), self.abs_cwd)

    def run(self):
        if (not os.path.isfile(os.path.join(self.abs_cwd, self.jar))) or (not self.jar.endswith('.jar')):
            raise OSError('{} is not a jar file.'.format(self.jar))

        if os.path.exists(os.path.join(self.abs_cwd, 'temp-log.txt')):
           self._log_file_name = 'temp-log-new.txt'

        log_file = open(os.path.join(self.abs_cwd, self._log_file_name), 'w')

        self._server = subprocess.Popen(
            ' '.join(['java', f'-Xms{self.min_RAM}', f'-Xmx{self.max_RAM}', '-jar', self.jar, 'nogui']),
            cwd=self.abs_cwd,
            shell=True,
            stdout=log_file,
            stderr=subprocess.PIPE,
            stdin=log_file
        )
        self.online = True

        while self.online:
            self.parser.process_events(self.new_events)
            time.sleep(self.update_wait)

    @property
    def new_events(self):
        with open(os.path.join(self.abs_cwd, 'temp-log.txt'), 'r') as log:
            events = log.readlines()
        if len(events) > self._old_events:
            new_events = events[self._old_events:]
            print(len(new_events), self._old_events, len(events))
            self._old_events += len(events) - self._old_events
            return new_events
        else:
            return []

    def _exec_cmd(self, cmd, *params):
        if not self.online:
            raise OSError('Server isn\'t started yet.')

        stdout, stderr = self._server.communicate(' '.join([cmd, *params]))

        return stdout, stderr

    def killserver(self):
        self._server.kill()

    # Commands
    def run_cmd(self, *args):
        return self._exec_cmd(*args)

    def op(self):
        pass

    # Server Folder Analysis
    @property
    def properties(self):
        with open(os.path.join(self.abs_cwd, 'server.properties'), 'r') as file:
            lines = file.readlines()
        properties = {}
        for line in lines:
            if not line.startswith('#'):
                k, v = line.split('=')
                properties[k] = v
        return properties

    @properties.setter
    def properties(self, value: dict):
        with open(os.path.join(self.abs_cwd, 'server.properties'), 'w') as file:
            properties = ['='.join(item) for item in value.items()]
            file.writelines(properties)

    @property
    def banned_ips(self):
        with open(os.path.join(self.abs_cwd, 'banned-ips.json'), 'r') as file:
            banned_ips = json.load(file)
        return banned_ips

    @property
    def banned_players(self):
        with open(os.path.join(self.abs_cwd, 'banned-players.json'), 'r') as file:
            banned_players = json.load(file)
        return banned_players

    @property
    def ops(self):
        with open(os.path.join(self.abs_cwd, 'ops.json'), 'r') as file:
            ops = json.load(file)
        return ops



import datetime
import requests
import json
from typing import Union


class Player:
    cache = {}

    def __init__(self, player=None, uuid=None):
        if player and uuid:
            pass
        elif uuid:
            player = self.get_player(uuid)
        elif player:
            uuid = self.get_uuid(player)
        else:
            raise ValueError('class Player needs either a player username or uuid to initialize.')
        self.player = player
        self.uuid = uuid

    def get_player(self, uuid):
        if uuid in self.cache.values():
            return {v:k for k,v in self.cache.items()}[uuid]
        api = 'https://api.mojang.com/user/profiles/%s/names' % uuid
        try:
            response = requests.get(api).json()
            if 'error' in response and 'errorMessage' in response:
                ResponseError = type(response['error'], tuple([Exception]), {})
                raise ResponseError(response['errorMessage'])
            self.cache[response[-1]['name']] = uuid
            return response[-1]['name']
        except json.decoder.JSONDecodeError:
            raise ValueError('No players exist with uuid, {}, according to Mojang.'.format(uuid))

    def get_uuid(self, player):
        if player in self.cache:
            return self.cache[player]

        api = 'https://api.mojang.com/users/profiles/minecraft/%s' % player
        try:
            response = requests.get(api).json()
            if 'error' in response and 'errorMessage' in response:
                ResponseError = type(response['error'], tuple([Exception]), {})
                raise ResponseError(response['errorMessage'])
            self.cache[player] = response['id']
            return response['id']
        except json.decoder.JSONDecodeError:
            raise ValueError('Player {} Doesn\'t exist, according to Mojang.'.format(player))

    def __repr__(self):
        return '<Player {}:{}>'.format(self.player, self.uuid)


class Event:
    def __init__(self, raw_event: str):
        self.raw = raw_event
        if self.raw.startswith('['):
            info = self.raw.replace('\n', '').split(': ')[0]
            message = self.raw.replace('\n', '').split(': ')[1:]
            self.message = ': '.join(message).replace('[', ' ').replace(']', '')
            time, self.process = info.replace('[', '').replace(']', '').split(' ')[0], ' '.join(info.replace('[', '').replace(']', '').split(' ')[1:])
            now = datetime.datetime.now()
            self.time = datetime.datetime(now.year, now.month, now.day, *[int(num) for num in time.split(':')])
        else:
            raise ValueError('Not a recognized event,', self.raw)

        self.type = self._event_type()

    def _event_type(self) -> Union[str, None]:
        msg_comps = self.message.split(' ')
        try:
            Player(msg_comps[0])  # Check if valid player from Mojang.
            if msg_comps[1] == 'left':
                return 'player_leave'
            elif msg_comps[1] == 'joined':
                return 'player_join'
            elif ' '.join(msg_comps[1:3]) == 'lost connection':
                return 'player_disconnect'
            elif msg_comps[1].startswith('/'):
                return 'player_connect'
        except ValueError:
            print(' '.join(msg_comps[0:3]), 'Preparing spawn area')
            if ' '.join(msg_comps[0:3]) == 'Preparing spawn area':
                return 'spawn_prep'
            if msg_comps[0] == 'Done':
                return 'ready'
            if ' '.join(msg_comps[0:3]) == 'Can\'t keep up!':
                return 'ticks_behind'
            if ' '.join(msg_comps[0:6]) == 'Considering it to be crashed,':
                return 'crash'
            if msg_comps[0].startswith('<') and msg_comps[0].endswith('>'):
                try:
                    Player(msg_comps[0].replace('<', '').replace('>', ''))
                    return 'player_msg'
                except ValueError:
                    return str(None)

    @property
    def ctx(self):
        return [self.time, self._parse_message(self.message)]

    def _parse_message(self, message: str):
        pass


class Parser:
    def process_events(self, events):
        for raw_event in events:
            try:
                event = Event(raw_event)
                for attr in self.__dict__:
                    if attr == 'on_' + event.type:
                        self.__dict__[attr](*event.ctx)

            except Exception as err:
                self.on_parsing_error(err, raw_event)

    def on_ready(self, time, startup_time):
        print('ready')

    def on_spawn_prep(self, completion_percent: float):
        print('prep')

    def on_player_join(self, time, player):
        print('join')

    def on_player_connect(self, time, player):
        print('connect')

    def on_player_leave(self, time, player):
        print('leave')

    def on_player_disconnect(self, time, player):
        print('disconnect')

    def on_ticks_behind(self, time, ticks):
        print('behind')

    def on_command(self, time, command):
        print('command')

    def on_player_message(self, time, player, message):
        print('message')

    def on_crash(self):
        print('crashed')

    def on_parsing_error(self, err, raw_event):
        print(err.__class__.__name__, err, raw_event)


from setuptools import setup

setup(
    name='MCServerInterface',
    version='0.0.2',
    packages=['mcserver'],
    url='https://github.com/TeamNightSky/MCServerWrapper',
    license='GNU License',
    author='Pixymon',
    author_email='nlarsen23.student@gmail.com',
    description='Integratable API for Local Minecraft Servers',
    install_requires=['requests']
)
